/*
 * sprite.c
 *
 * Last modified: <2008/06/27 19:55:35 +0900> By SUMIDA
 */
#include <stdlib.h>
#include <search.h>
#include <math.h>
#include <SDL.h>
#include "system.h"
#include "sprite.h"
#include "surface.h"

static int sprite_count = 0;
static SPRITE* sprite_table[SPRITE_MAX];

void sprite_init(void) {

	printf("sprite_init ... ");
	sprite_count = 0;
	memset(sprite_table, NULL, sizeof(sprite_table));
	printf("done.\n");
}

void sprite_clean(void) {
	sprite_count = 0;
}

void sprite_put(SPRITE* sprite) {

	if (sprite_count < SPRITE_MAX) {
		//printf("sprite_put: surface_id=%d, frame_no=%d\n", sprite->surface_id, sprite->frame_no);
		sprite_table[sprite_count++] = sprite;
	}
}

static int sprite_comp(const void* p0, const void* p1) {

	int pri0 = (*(SPRITE**)p0)->options & SPRITE_MSK_PRIORITY;
	int pri1 = (*(SPRITE**)p1)->options & SPRITE_MSK_PRIORITY;

	return pri0 - pri1;
}

void sprite_rendering(SDL_Surface* screen) {

	if (sprite_count > 0) {
		int count = sprite_count;

		qsort(sprite_table, count, sizeof(SPRITE*), sprite_comp);

		while (count-- > 0) {
			sprite_draw(screen, sprite_table[count]);
		}
	}
}

void sprite_draw(SDL_Surface* screen, SPRITE* sprite) {

	if (sprite->options & SPRITE_VISIBLE) {

		int func_id = (sprite->options & SPRITE_MSK_FRAME) >> SPRITE_SFT_FRAME;
		FUNC_SURFACE_DRAW draw = func_surface_draw[func_id];

		int x = RealToInt(sprite->x);
		int y = RealToInt(sprite->y);

		if (surface_is_exists(sprite->surface_id) != SDL_TRUE) {
			fprintf(stderr, "Surface is nothing. (%d)\n", sprite->surface_id);
			return;
		}
/*
		if (func_id >= FUNCID_SURFACE_DRAW_MAX) {
			fprintf(stderr, "Draw function is nothing. (%d)\n", func_id);
			return;
		}
*/
//		printf("func_id=%d\n", func_id);
/*
		if (sprite->options & SPRITE_USE_ALPHA) {

			surface_set_alpha(
				sprite->surface_id,
				(sprite->options & SPRITE_MSK_ALPHA) >> SPRITE_SFT_ALPHA);
			draw(screen, sprite->surface_id, sprite->frame_no, x, y);
			surface_reset_alpha(sprite->surface_id);

		} else {
			printf("surface_id=%d, frame_no=%d\n", sprite->surface_id, sprite->frame_no);
			draw(screen, sprite->surface_id, sprite->frame_no, x, y);
		}
*/
		if (draw == NULL) {
			surface_draw(screen, sprite->surface_id, x, y);
		} else {
			draw(screen, sprite->surface_id, sprite->frame_no, x, y);
		}
	}
}

void sprite_action(SPRITE *sprite) {

	if (sprite->options & SPRITE_USE_MOVE) {

		if (++sprite->move_counter > sprite->move_interval) {
			sprite->move_counter = 0;
			sprite->x += sprite->dx;
			sprite->y += sprite->dy;
		}
	}

	if (sprite->options & SPRITE_USE_ANIME
		&& !(sprite->options & SPRITE_STOP_ANIME)) {

		if (++sprite->frame_counter > sprite->frame_interval) {
			sprite->frame_counter = 0;
			if (++sprite->frame_no > sprite->frame_end) {
				if (sprite->options & SPRITE_LOOP_ANIME) {
					sprite->frame_no = sprite->frame_start;
				} else {
					sprite->frame_no = sprite->frame_end;
					sprite->options |= SPRITE_STOP_ANIME;
				}
			}
		}
	}

	if (sprite->options & SPRITE_USE_LOWER) {

		Uint32 pri = sprite->options & SPRITE_MSK_PRIORITY;

		if (pri < SPRITE_MSK_PRIORITY) {
			sprite->options &= ~SPRITE_MSK_PRIORITY;
			sprite->options |= pri + 1;
		}
	}
}

SDL_bool sprite_collision_check(SPRITE* s0, SPRITE* s1) {

	SDL_bool result = SDL_FALSE;
#if 0
	printf("s0->options=0x%X, s1->options=0x%X\n", s0->options, s1->options);
#endif
	if (s0->options & SPRITE_USE_COLLISION
		&& s1->options & SPRITE_USE_COLLISION) {
		RealRect rect;

		rect.x0 = s1->x + s1->collision.x0 - s0->x;
		rect.y0 = s1->y + s1->collision.y0 - s0->y;
		rect.x1 = s1->x + s1->collision.x1 - s0->x;
		rect.y1 = s1->y + s1->collision.y1 - s0->y;
/*
		printf("s0: x0=%d, y0=%d, x1=%d, y1=%d\n", s0->collision.x0, s0->collision.y0, s0->collision.x1, s0->collision.y1);
		printf("s1: x0=%d, y0=%d, x1=%d, y1=%d\n", rect.x0, rect.y0, rect.x1, rect.y1);
*/
		if (rect.x0 < s0->collision.x1 && rect.x1 > s0->collision.x0
			&& rect.y0 < s0->collision.y1 && rect.y1 > s0->collision.y0) {

			/*printf("hit\n");*/
			result = SDL_TRUE;
		}
	}

	return result;
}

double sprite_get_angle(SPRITE *s0, SPRITE *s1) {

	SURFACE *surface0 = surface_get(s0->surface_id);
	SURFACE *surface1 = surface_get(s1->surface_id);

	double dx = (s1->x + (IntToReal(surface1->frame_width) >> 1)
				 - s0->x - (IntToReal(surface0->frame_width) >> 1)) / 65536.0;
	double dy = (s1->y + (IntToReal(surface1->frame_height) >> 1)
				 - s0->y - (IntToReal(surface0->frame_height) >> 1)) / 65536.0;

	return angle(dx, dy);
}

void sprite_set_angle(SPRITE *s, double angle, Real speed) {

	double r = angle * M_PI / 180;

	s->dx = (Real)(sin(r) * speed);
	s->dy = (Real)(cos(r) * speed);

	if (angle >= 180) {
		s->dx = -s->dx;
		s->dy = -s->dy;
	}
	/*REPORT(format("angle=%f, x=%d, y=%d", angle, s->dx, s->dy));*/
}

void sprite_fire(SPRITE* sprite, SPRITE *parent, Real ox, Real oy) {

	SURFACE *surface_this = surface_get(sprite->surface_id);
	SURFACE *surface_parent = surface_get(parent->surface_id);

	sprite->x = parent->x + ox
		+ (IntToReal(surface_parent->frame_width - surface_this->frame_width) >> 1);
	sprite->y = parent->y + oy
		+ (IntToReal(surface_parent->frame_height - surface_this->frame_height) >> 1);
}

void sprite_homing(SPRITE* sprite, SPRITE* target, Real speed) {

	SURFACE *surface_this = surface_get(sprite->surface_id);
	SURFACE *surface_target = surface_get(target->surface_id);
	double dx = (target->x + (IntToReal(surface_target->frame_width) >> 1)
				 - sprite->x - (IntToReal(surface_this->frame_width) >> 1)) / 65536.0;
	double dy = (target->y + (IntToReal(surface_target->frame_height) >> 1)
				 - sprite->y - (IntToReal(surface_this->frame_height) >> 1)) / 65536.0;
	double d = sqrt(dx * dx + dy * dy);

	if (d) {
		double spd = speed / 65536.0;
		sprite->dx = (Real)((dx / d * spd) * 65536);
		sprite->dy = (Real)((dy / d * spd) * 65536);
	} else {
		sprite->dx = 0;
		sprite->dy = speed;
	}
}

